#/usr/bin/python

'''The go pane from the main notebook.

This is where we execute all the processes that were configured in all
the other panes of the notebook.
'''

import gtk
import config
import Model
import cropper
import rotater
from generator_task import GeneratorTask
from oneline import comm
from NotebookPane import NotebookPane
from rotate_widget import RotateWidget
from rotate_widget import EdgeSelectCanvas
from rotate_widget import degrees_from_edge
from rotate_widget import edge_from_degrees
from GUI.misc import set_lower

def user_rotate_page_if_undone(page):
    '''Calls user_rotate_page if this page isn't already rotated.
    Returns what user_rotate_page returns (or would return if we had
    called it).'''

    try:
        if page.blank:
            return EdgeSelectCanvas.BLANK
        if page.rotate_method == Model.USER:
            return edge_from_degrees(page.rotate_degrees)
    except:
        pass

    return user_rotate_page(page)


def user_rotate_page(page):
    '''Pop up a dialog and let user choose rotation.  Record the
    result.  This function does not actually DO the rotation.  We
    leave that for the bulk processing stage.

    Returns result of user pick: EdgeSelectCanvas.TOP | EdgeSelectCanvas.LEFT |
    EdgeSelectCanvas.RIGHT | EdgeSelectCanvas.BOTTOM |
    EdgeSelectCanvas.BLANK | gtk.RESPONSE_REJECT

    gtk.RESPONSE_REJECT means user hit cancel or hit OK before making
    a selection.  In the case of user confusion of this kind, we
    should just move on.  Give them a different page.  Come back to
    this page later.

    TODO: should we reuse the widget?  Might be faster.'''
        
    ## Let user pick a side
    rw = RotateWidget(page.path)
    rw.show_all()
    run = rw.run()
    rw.destroy()

    if run == gtk.RESPONSE_REJECT:
        return run
    if run == EdgeSelectCanvas.BLANK:
        page.blank = True
        return run

    ## Process edge selection
    page.rotate_degrees = degrees_from_edge(run)
    page.rotate_method = Model.USER
    page.rotate_confidence = Model.USER_SPECIFIED
    
    return run

class GoPane(NotebookPane):
    'Select which of the steps to do and then GO!'
    def __init__(self, M):
        NotebookPane.__init__(self)
        self.make_tab('_GO!', gtk.STOCK_EXECUTE)
        self.create_buttons()
        self.M = M

    def create_buttons(self):
        'Populate pane with option boxes.'

        self.content = gtk.VBox()
        label = gtk.Label('Select processes to run on the book, then hit GO!')
        self.content.pack_start(label, False)
        
        ## checkboxes
        self.rotate_btn = gtk.CheckButton(label='Rotate')
        self.crop_btn = gtk.CheckButton(label='Crop')
        self.ocr_btn = gtk.CheckButton(label='OCR (not implemented)')
        self.ocr_btn.set_sensitive(False)
        self.output_btn = gtk.CheckButton(label='Formatted Output (not implemented)')
        self.output_btn.set_sensitive(False)
        self.content.pack_start(self.rotate_btn, False)
        self.content.pack_start(self.crop_btn, False)
        self.content.pack_start(self.ocr_btn, False)
        self.content.pack_start(self.output_btn, False)
        
        ## Go Button
        self.go_btn = gtk.Button(stock="gtk-execute")
        self.go_btn.set_sensitive(False)
        self.go_btn.connect('clicked', self.go_clicked)
        label = self.go_btn.get_children()[0]
        label = label.get_children()[0].get_children()[1]
        label = label.set_label("GO!")
        self.content.pack_start(self.go_btn, False)
        self.add_with_viewport(self.content)

        ## Toggling checkboxes activates go button
        self.rotate_btn.connect('toggled', self.set_go_button_sensitive)
        self.crop_btn.connect('toggled', self.set_go_button_sensitive)
        self.ocr_btn.connect('toggled', self.set_go_button_sensitive)
        self.output_btn.connect('toggled', self.set_go_button_sensitive)

    def set_modal_parent_func(self, func):
        '''Func is the function we call when we want to lock the user
        out of the other tabs.'''
        self.modal_parent_func = func

    def set_modal(self, modal):
        '''(De)activate all the tabs except Go!'''
        self.modal_parent_func(not modal)

    def set_go_button_sensitive(self, button):
        '''Toggles GO! sensitivity based on whether user has chosen
        any processing paths.'''
        self.go_btn.set_sensitive(
            self.rotate_btn.get_active() or
            self.crop_btn.get_active() or
            self.ocr_btn.get_active() or
            self.output_btn.get_active())

    def set_lower(self, widget):
        '''Set the lower part of the pane to the widget.'''
        if not hasattr(self, 'lower'):
            self.lower = gtk.VBox()        
        set_lower(self, widget, self.lower, self.content)

    def map_cb(self, action):
        'Called when this pane is chosen and becomes visible.'

        config.help_url='help:go'

        ## set rotate button based on group ignore/skip stuff
        ignore = True
        for g in self.M.files:
            if self.M.files[g] and self.M.rotate.group[g].mode != Model.IGNORE:
                ignore = False
        self.rotate_btn.set_active(not ignore)            
        self.rotate_btn.set_sensitive(not ignore)

        ## set crop button based on group ignore/skip stuff
        ignore = True
        for g in self.M.files:
            if self.M.files[g] and self.M.crop.group[g].mode != Model.IGNORE:
                ignore = False
        self.crop_btn.set_active(not ignore)
        self.crop_btn.set_sensitive(not ignore)


    def guided_rotate(self, group_id):
        '''Do guided rotation on group identified by group_id.

        Pop up the rotate widget a bunch of times, collect user
        input and deduce rotation from the pattern of the input.'''

        ## We recognize three patterns: all the same, alternating, and
        ## all one way followed by all 180 degrees the other way.
        ## Look for those patterns.  If we find the pattern, good.  If
        ## not, go to automatic rotating.  

        ## We assume the pages in M.files are sorted by filename.
        ## That plus the assumption that pages were scanned in order
        ## allows us to assume that consecutive files are somewhat
        ## consecutive in the book too.  Somewhat because it could be
        ## every-other-page.  Note that we do alphanumeric sort rather
        ## than asciibetical sort, so page2.png will come before
        ## page10.png.

        ## Most of the time, every page in the group
        ## will require the same rotation.


        #Let user hand-rotate first 3 pages
        p = pages_rotated = 0
        all_the_same = every_other_180 = half_one_half_other = True
        flip = False
        last = flip_dir = flop_dir = None
        while pages_rotated < 3 and p < len(self.M.files[group_id]):
            resp = user_rotate_page_if_undone(self.M.pages[self.M.files[group_id][p]])

            if resp not in [gtk.RESPONSE_REJECT, EdgeSelectCanvas.BLANK]:
                pages_rotated = pages_rotated + 1
                if last and resp != last:
                    all_the_same = False
                last = resp

                ## Check if every other page is the same rotation
                if flip and flip_dir and flip_dir != resp: every_other_180 = False
                if flip and not flip_dir: flip_dir = resp
                if not flip and flop_dir and flop_dir != resp: every_other_180 = False
                if not flip and not flop_dir: flop_dir = resp

                ## If user said they're all the same, just go with it
                if self.M.rotate.group[group_id].consistent:
                    self.M.rotate.group[group_id].pattern = Model.ALL_SAME
                    return

            flip = not flip
            p = p + 1

        # Make sure flip and flop are 180 off from each other
        if ((degrees_from_edge(flip_dir) + degrees_from_edge(flop_dir)) % 180) != 0:
            every_other_180 = False

        last2 = flip_dir = flop_dir = None

        #Let user hand-rotate last 3 pages
        p = len(self.M.files[group_id]) - 1
        while pages_rotated < 6 and p >= 0:
            resp = user_rotate_page_if_undone(self.M.pages[self.M.files[group_id][p]])
            if resp not in [EdgeSelectCanvas.BLANK, gtk.RESPONSE_REJECT]:
                pages_rotated = pages_rotated + 1
                if last and resp != last:
                    all_the_same = False
                last = resp

                if last2 and resp !=last2:
                    half_one_half_other = False
                last2 = resp
                                 
                ## Check if every other page is the same rotation
                if flip and flip_dir and flip_dir != resp: every_other_180 = False
                if flip and not flip_dir: flip_dir = resp
                if not flip and flop_dir and flop_dir != resp: every_other_180 = False
                if not flip and not flop_dir: flop_dir = resp

            if pages_rotated == 4 and all_the_same:
                self.M.rotate.group[group_id].pattern = Model.ALL_SAME
                return

            flip = not flip
            p = p - 1

        ## Make sure flip and flop are 180 off from each other
        if ((degrees_from_edge(flip_dir) + degrees_from_edge(flop_dir)) % 180) != 0:
            every_other_180 = False

        if all_the_same:
            self.M.rotate.group[group_id].pattern = Model.ALL_SAME
        elif half_one_half_other:
            self.M.rotate.group[group_id].pattern = Model.HALF_AND_HALF
        elif every_other_180:
            self.M.rotate.group[group_id].pattern = Model.ALTERNATING
        else:
            self.M.rotate.group[group_id].pattern = Model.AUTO


    def collect_user_guidance(self, M):
        '''Let the user rotate some images if needed.  Set
        M.rotate.group[group_id].pattern so we know what kind of
        pattern to use in the rotation.'''

        ## Collect rotate data from user input
        if self.rotate_btn.get_active():
            for g in M.files:
                ##if M.files[g] and M.rotate.group[g].mode != Model.IGNORE:
                if M.rotate.group[g].mode != Model.IGNORE:
                    if (M.rotate.group[g].consistent or
                        M.rotate.group[g].mode == Model.GUIDED):
                        self.guided_rotate(g)


    def do_rotation(self):
        '''Start a thread to do the rotation of all the groups.  Move
        on to crop when finished.'''
        if self.rotate_btn.get_active():
            self.thread = GeneratorTask(rotater.rotate_groups, 
                                        self.progress.set_fraction,
                                        self.do_crop)
            self.thread.start(self.M, self.update_group, self.progress.set_text)
        else:
            # Rotate as if mode is IGNORE
            for g in self.M.files:
                rotater.rotate_ignored_group(self.M, g)

            # Now crop
            self.do_crop()

    def update_group(self, txt):
        '''Update user as to progress by setting the group_label and using comm.out'''
        self.group_label.set_text(txt)
        comm.out(txt)

    def do_crop(self):
        '''Start a thread to crop all the groups.'''
        if self.crop_btn.get_active():
            self.thread = GeneratorTask(cropper.crop_groups,
                                        self.progress.set_fraction,
                                        self.do_crop_correct)
            self.thread.start(self.M, self.update_group, self.progress.set_text)
        else:
            self.cancel_thread()

    def do_crop_correct(self):
        '''Apply statistical correction to the cropped images.'''
        sanity = cropper.SanityCheck(self.M)
        self.thread = GeneratorTask(sanity.do_sanity_check,
                                    self.progress.set_fraction,
                                    self.cancel_thread)
        self.thread.start(self.update_group, self.progress.set_text)

    def cancel_thread(self, button=None):
        try:
            self.thread.stop()
        except:
            pass
        self.lower.destroy()
        self.go_btn.set_sensitive(True)
        self.set_modal(False)
        comm.out("Done processing.")

    def go_clicked(self, button):
        '''This is where we kick off the processing that was
        configured in all the prior panes.

        This func kicks off a thread to do the rotation, crop, etc.
        We need the new thread because otherwise control never comes
        back around to the main gtk loop, which means the display
        never gets updated.  The GUI gets real ugly when that happens.

        TODO: we should be able to skip pages that were rotated by prior runs
        TODO: we need to update M.pages as we rotate'''

        self.collect_user_guidance(self.M)
        self.go_btn.set_sensitive(False)
        self.set_modal(True)

        box = gtk.VBox()
        hbox = gtk.HBox()
        self.group_label = gtk.Label('')
        hbox.pack_start(self.group_label, False, False, padding = 2)
        box.pack_start(hbox, False, False, padding = 5)
        self.progress = gtk.ProgressBar(adjustment=None)
        box.pack_start(self.progress, False, padding = 5)
        self.cancel_button = gtk.Button(stock = gtk.STOCK_CANCEL)
        self.cancel_button.connect("clicked", self.cancel_thread)
        box.pack_start(self.cancel_button, False, padding = 5)
        self.set_lower(box)
        self.progress.show()
        self.do_rotation()
