#!/usr/bin/python

'''Some panes in the main notebook contain sub groups.  This file
contains classes to support such panes.

The nomenclature here is confusing, but we provide two classes below.
There is GroupNotebookPane and GroupSubPane.  GroupNotebookPane is the
pane of the main tab that contains sub panes for each group.
GroupSubPane is the sub panes.

Anybody who comes up with better and clearer names for these classes
wins a prize.'''

import gtk
import re
import config
from NotebookPane import NotebookPane
from GUI.misc import OkCancelDlg
from oneline import comm

class GroupSubPane(NotebookPane):
    '''Parent class for RotateGroupPane and CropGroupPane

    Put things in self.content.  That way we can update the widget by
    destroying and replacing self.content.
    '''
    def __init__(self, M, id):
        NotebookPane.__init__(self)
        self.id = id  # this is the id string of the group
        self.make_tab(M, id)
        self.set_policy(gtk.POLICY_NEVER,gtk.POLICY_AUTOMATIC)
        self.content = gtk.VBox()

    def make_tab(self, M, id):
        '''Provide a widget that we can pass to the notebook when appending a
        new tab and pane'''
        f = M.pages[M.files[id][0]].path.partition(M.path)
        # We have to reverse the string (hence [::-1]) because re
        # doesn't do right-to-left search
        f = re.sub('\d+',config.fname_token, f[2][::-1], 1)[::-1]
        f = re.sub('\A/', '', f)

        NotebookPane.make_tab(self, f, gtk.STOCK_CLOSE, True)
        self.tab_btn.connect('clicked', self.destroy_cb, M, id)

    def destroy_cb(self, action, M, id):
        '''Remove a group from the files list.

        M is our data model and id is the group to remove.
        '''

        result = OkCancelDlg(
            "Closing this tab will remove this group of files from "
            "processing.  They won't get rotated, cropped or "
            "otherwise processed in any way and won't appear "
            "in the output.\n\n"
            "Press OK to remove this group of files.").result
        if (result == gtk.RESPONSE_CANCEL or 
            result == gtk.RESPONSE_DELETE_EVENT):
            return False

        comm.debug("Removing group %s" % id)

        M.remove_group(id)
        self.destroy()

class GroupNotebook(gtk.Notebook):
    'If the images to rotate/crop contain multiple groups, put them in a notebook.'
    def __init__(self):
        gtk.Notebook.__init__(self)
        self.set_scrollable(True)
        self.set_show_border(False)

class GroupNotebookPane(NotebookPane):
    '''Parent class for group panes in the rotate and crop sections'''
    def __init__(self, M, panetype):
        '''Make group notebook and sub-panes for rotate and crop
        panes.
        
        M is the data model.
        panetype is a ref to the init function of the subpane (either RotatePane or CropPane)

        Constructs a GroupNotebook (at self.notebook) containing a
        pane for each group in M.files.  Updates the rotate or crop
        pane to display this new notebook.'''
        NotebookPane.__init__(self)
        self.content = gtk.VBox()
        self.notebook = GroupNotebook()

        page_num=0
        #self.group_widget = []
        for g in sorted(M.files):
            #M.add_group(g)

            widget = panetype(M, g)
            widget.connect("destroy", self.destroy_cb, M, page_num)
            pane = gtk.VBox()
            pane.visible = True
            pane.pack_start(widget, True)
            self.notebook.append_page(pane, widget.tab)
            self.notebook.set_tab_label_packing(widget, True, False, gtk.PACK_END)
            page_num += 1

        if page_num == 1:
            self.notebook.set_show_tabs(False)

        self.content.pack_start(self.notebook, True)
        self.add_with_viewport(self.content)        

    def destroy_cb(self, act, M, page_num, propagate = True):
        '''Remove GUI bookkeeping associated with the destroyed pane.
        Does not delete rotate and group instances from the model.
        '''

        #comm.debug("GNP.destroy_cb: Deleting page %s" % page_num)

        self.notebook.get_nth_page(page_num).hide()
        self.notebook.get_nth_page(page_num).visible = False

        ## Count open tabs, hide tabs if they're not needed anymore
        count_open_tabs = 0
        for i in range(self.notebook.get_n_pages()):
            if self.notebook.get_nth_page(i).visible:
                count_open_tabs += 1
        if count_open_tabs == 1:
            self.notebook.set_show_tabs(False)

        ## If destroy starts in rotate, call crop and vice versa
        if act:
            try:
                if act.__class__.__name__ == 'RotateSubPane':
                    self.parent.crop.destroy_cb(None, M, page_num)
                else:
                    self.parent.rotate.destroy_cb(None, M, page_num)
            except:
                pass






#    def files_loaded(self, M, panetype):
#        '''Make group sub-panes for rotate and crop panes.
#        
#        M is the data model.
#        panetype is a ref to the init function of the subpane (either RotatePane or CropPane)
#
#        Constructs a GroupNotebook (at self.notebook) containing a
#        pane for each group in M.files.  Updates the rotate or crop
#        pane to display this new notebook.
#
#        This func is no longer called.  We can delete it soon.'''
#
#        comm.debug("GroupPane.files_loaded")
#        #self.content.destroy()
#        self.content = gtk.VBox()
#        self.notebook = GroupNotebook()
#
#        page_num=0
#        self.group_widget = []
#        for g in sorted(M.files):
#            M.add_group(g)
#
#            pane = gtk.VBox()
#
#            widget = panetype(M, g)
#            widget.connect("destroy", self._destroy_cb, M, page_num)
#
#            pane.pack_start(widget, True)
#            self.group_widget.append(pane)
#            self.notebook.append_page(pane, widget.tab)
#            self.notebook.set_tab_label_packing(widget, True, False, gtk.PACK_END)
#            page_num = page_num + 1
#
#        if page_num == 1:
#            self.notebook.set_show_tabs(False)
#
#        self.content.pack_start(self.notebook, True)
#        self.add_with_viewport(self.content)
        
