#!/usr/bin/python

'''Module contains classes and functions for cropping individual files
as well as doing some statistical analysis of those files to do sanity
checks.'''

import numpy
from oneline import comm
import config
import graphic
from misc import *
import Model
import cli

class CropMisc(object):
    '''Parent class containing crop functions shared by Cropper and SanityCheck
    classes.'''

    def __init__(self):
        pass

    def black(self,color, **kwargs):
        '''returns True if color is black

        Keyword arg black_max is the maximum level for a pixel to be
        considered black 

        TODO: look at the histogram to determine black and white thresholds'''
        kwargs = set_defaults( {'black_max': 256/3}, kwargs )

        return color <= kwargs['black_max']

    def white(self, color, **kwargs):
        '''returns True if color is white

        Keyword arg white_min is the minimum level for a pixel to be
        considered white.

        TODO: look at the histogram to determine black and white thresholds'''

        kwargs = set_defaults( {'white_min': 256/4}, kwargs )
        
        return color >= kwargs['white_min']

    def next_white_row(self, left, top, right, bottom, dir, streak, **kwargs):
        '''Step through the horizontal lines until we have streak lines in a
        row that are white.

        Returns the first white line.

        dir == 'up' or 'down'
        '''
 
        if dir == 'down':
            t = top
            stop = bottom - 1
        elif dir == 'up':
            t = bottom - 1
            stop = top
        else:
            assert False, "dir parameter must be 'up' or 'down'"

        found_black = True
        curr_streak = 0
        while t != stop and curr_streak < streak:
            x = left
            found_black=False
            while x < right and not found_black:
                if not self.white(self.im.getpixel((x,t)), **kwargs):
                    found_black = True
                x = x + 3 

            curr_streak = (curr_streak + 1) * (not found_black)
            t = t + ((dir == 'down') * 2 - 1)

        return t - streak  * ((dir == 'down') * 2 - 1)

    def next_black_row(self, left, top, right, bottom, dir, streak, **kwargs):
        '''Step through the horizontal lines, until we have streak lines in a
        row containing black pixels.  Return the last line that was all white.

        dir == 'up' or 'down'
        '''

        if dir == 'down':
            t = top
            stop = bottom - 1
        elif dir == 'up':
            t = bottom - 1
            stop = top
        else:
            assert False, "dir parameter must be 'up' or 'down'"

        found_black = False
        curr_streak = 0
        while t != stop and curr_streak < streak:
            x = left
            found_black=False
            while x < right and not found_black:
                if self.black(self.im.getpixel((x,t)), **kwargs):
                    found_black = True
                x = x + 3 

            curr_streak = (curr_streak + 1) * found_black
            t = t + ((dir == 'down') * 2 - 1)

        return t - streak * ((dir == 'down') * 2 - 1)

    def lower_the_top(self, left, top, right, bottom, **kwargs):
        '''Return line from the top where we can crop so we only get
        whitespace up there.'''

        comm.verbose("   Lowering the top.")

        t1 = self.next_white_row(left,top,right,bottom, 'down', 3)
        t2 = self.next_black_row(left, t1+1, right, bottom, 'down', 5)

        ## Breathing room
        if t2 - t1 < 5:
            return t1
        else:
            return t2 - 5

    def raise_the_bottom(self, left, top, right, bottom, **kwargs):
        '''Return line from the bottom where we can crop so we only get
        whitespace down there.'''

        comm.verbose("   Raising the bottom.")

        t1 = self.next_white_row(left,top,right,bottom, 'up', 3)
        t2 = self.next_black_row(left, top, right, t1-1, 'up', 5)

        ## Breathing room
        if t1 - t2 < 5:
            return t1
        else:
            return t2 + 5

    def next_white_col(self, left, top, right, bottom, 
                       start, dir, streak, **kwargs):
        '''Step through the vertical lines until we have streak consecutive
        cols that are white. Returns the first white line.

        dir == 'left' or 'right'
        start is the column to start at.
        '''

        x = start
        if x == self.im.w:
            x = x - 1
        stop = (self.im.w - 1) * (dir == 'right')
        assert (dir == 'left' or 
                dir == 'right'), "dir parameter must be 'left' or 'right'"

        found_black = True
        curr_streak = 0
        while x != stop and curr_streak < streak:
            y = top
            found_black=False
            while y < bottom and not found_black:
                if not self.white(self.im.getpixel((x,y)), **kwargs):
                    found_black = True
                y = y + 2

            curr_streak = (curr_streak + 1) * (not found_black)
            x = x + ((dir == 'right') * 2 - 1)

        return x - streak  * ((dir == 'right') * 2 - 1)

    def breathing_room_sides(self, l, r):
        '''Adjust crop outward a few pixels so we don't confuse the OCR'''

        l = l - 5
        if l < 0:
            l = 0
        r = r + 5
        if r > self.im.w-1:
            r = self.im.w-1

        return l, r

    def tune_sides(self, left, top, right, bottom, **kwargs):
        '''Adjust left and right crop outward if they are not clean crops.'''

        comm.verbose("   Tuning left and right edges.")

        l = self.next_white_col(left,top,self.im.w,bottom, left, 'left', 5)
        r = self.next_white_col(left,top,self.im.w,bottom, right, 'right', 5)
        l, r = self.breathing_room_sides(l, r)
        return l, r


def crop(M, page_id, **opt):
    '''Make a cropper instance and crop one file.'''

    page = M.pages[page_id]

    comm.verbose("Cropping %s" % page.path)

    if (config.cont and
        os.access(page.crop_fname, os.F_OK) and 
        page.crop_fname in M.crop.lines):
        comm.debug("   Found %s. Skipping crop." % page.crop_fname)
        return

    c = Cropper(page.rotate_fname)
    g = c.crop(**opt)
    g.im.save(page.crop_fname)

    ## save crop lines for later analysis
    M.crop.lines[page.crop_fname] = {'fname':page.path, 'dest':page.crop_fname,
                                    'crop':g.crop}
    M.save()


def crop_groups(M, update_group = None, update_file = None):
    '''Auto-crop each group.

    update_group = a func that sends text to user.  We use it to
    update on which group we're on.

    update_file = similar to update_group.'''

    for g in M.files:

        ## skip closed tabs
        #if not M.files[g]:
        #    continue

        if update_group:
            update_group("Cropping %s" % g)
        group = M.crop.group[g]
        mode = group.mode
        if mode == Model.IGNORE:
            comm.verbose("Skip crop of %s" % g)
            for page_id in M.files[g]:
                page = M.pages[page_id]
                page.crop_fname = page.rotate_fname
        else:
            count = 1
            for page_id in M.files[g]:
                if update_file:
                    update_file(M.pages[page_id].path)
                yield count / (len(M.files[g]) * 1.0)
                crop(M, page_id)
                count = count + 1

class Cropper(CropMisc):
    def __init__(self, image=False):
        '''cropper(image)

        Class for cropping an image that contains a photo of a book page,
        with some whitespace in the margins and some material outside
        the pages of the book, as well as the page on the opposite
        leaf.

        image is an image file spec, a pils image or a graphic object.

        We asume that the book takes up at least 2/3 the frame in
        either dimension. TODO: Is this assumption still operative?

        Assumes the opposite leaf is the bottom edge (i.e. text will
        need to be rotated before OCRing)

        '''

        self.im = graphic.graphic(image)
        self.im.getdata()

        CropMisc.__init__(self)

    def find_whitest_segment(self, x, **kwargs):
        '''Iterate down the column and find the longest segment of
        whitespace.  Return the start and end y coordinate of that
        segment, lower number first.

        x is the column to look in.

        Kwargs

        There is an optional bug(?) in the algorithm in that if the
        longest white segment runs to the end of the area we are
        scanning, it doesn't get counted.  Sometimes that's useful,
        but most of the time we want to count segments from that area.  Set
        longest_can_be_last to True to do so.  The default is True.

        white_min is the minimum level for a pixel to be considered white
        black_max is the maximum level for a pixel to be considered black
        If white_min < black_man, pixels in between will be treated as the previous pixels.
        If white_min > black_max, pixels in the overlap will be treated as the previous pixels.

        y_range is the number of pixels to consider.  height/2 is the
        default.  If you change it, you might need to make
        corresponding changes in other procuedures in this class.

        y_skip is how far we advance the y scanner when considering
        pixels.  Set to 1 (the default) to consider every pixel in the
        pixels_to_scan range.  Set to 2 to scan every other one.  Set
        to three to look at every third.  Once you get beyond 5 or so,
        the scanning probably gets quite unreliable.

        '''



        kwargs = set_defaults( {'y_range': int(self.im.h / 2),
                                'longest_can_be_last' : True,
                                'y_skip' : 1},
                               kwargs)
                                     
        cur_white = True
        cur_start = 0
        long_start = 0
        long_end = 0
        y = 0

        while y < kwargs['y_range']:
            #pix = self.im.getpixel((x,y))
            pix = self.im.data[y * self.im.w + x]

            if cur_white:
                if self.black(pix):
                    if y - cur_start > long_end - long_start:
                        long_start = cur_start
                        long_end = y
                    cur_white = False
            elif self.white(pix):
                cur_white = True
                cur_start = y

            y = y + kwargs['y_skip']

        if kwargs['longest_can_be_last']:
            if cur_white:
                if y-cur_start > long_end - long_start:
                    long_start = cur_start
                    long_end = y

        return (long_start, long_end)

    def find_whitest_segment_horiz(self, y, **kwargs):
        '''Iterate across the row and find the longest segment of
        whitespace.  Return the start and end x coordinate of that
        segment, lower number first.

        y is the column to look in.

        We assume self.im.data has up-to-date info.

        Kwargs

        There is an optional bug(?) in the algorithm in that if the
        longest white segment runs to the end of the area we are
        scanning, it doesn't get counted.  Sometimes that's useful,
        but most of the time we want to count segments from that area.  Set
        longest_can_be_last to True to do so.  The default is True.

        white_min is the minimum level for a pixel to be considered white
        black_max is the maximum level for a pixel to be considered black
        If white_min < black_man, pixels in between will be treated as the previous pixels.
        If white_min > black_max, pixels in the overlap will be treated as the previous pixels.

        x_range is the number of pixels to consider.  width/2 is the
        default.  If you change it, you might need to make
        corresponding changes in other procedures in this class.

        x_skip is how far we advance the x scanner when considering
        pixels.  Set to 1 (the default) to consider every pixel in the
        pixels_to_scan range.  Set to 2 to scan every other one.  Set
        to three to look at every third.  Once you get beyond 5 or so,
        the scanning probably gets quite unreliable.
        '''



        kwargs = set_defaults( {'x_range': int(self.im.h / 2),
                                'longest_can_be_last' : True,
                                'x_skip' : 1},
                               kwargs)
                                     
        cur_white = True
        cur_start = 0
        long_start = 0
        long_end = 0
        x = 0

        while x < kwargs['x_range']:
            #pix = self.im.getpixel((x,y))
            pix = self.im.data[y * self.im.w + x]
            if cur_white:
                if self.black(pix):
                    if x - cur_start > long_end - long_start:
                        long_start = cur_start
                        long_end = x
                    cur_white = False
            elif self.white(pix):
                cur_white = True
                cur_start = x

            x = x + kwargs['x_skip']

        if kwargs['longest_can_be_last']:
            if cur_white:
                if y-cur_start > long_end - long_start:
                    long_start = cur_start
                    long_end = x

        return (long_start, long_end)


    def find_white_segments(self, **kwargs):
        '''Iterate across the width of the image and find the whitest segment
        of each vertical slice.

        Set kwargs[draw] = True to draw the white_segments and a
        horizontal crop line instead of doing the cropping.

        We assume self.im.data contains up-to-date self.im.getdata()
        info.  If it is missing, we grab it, but if it is wrong, we
        never find out.'''
        
        kwargs = set_defaults( {'draw' : False,
                                'y_skip' : 3}, kwargs)
        comm.verbose("   Finding white segments.")
        if kwargs['draw']:
            draw = ImageDraw.Draw(self.im.im)

        if not self.im.data:
            self.im.getdata()

        white_segments = []
        for col in range(self.im.w):
            y1, y2 = self.find_whitest_segment(col, **kwargs)
            white_segments.append((y1,y2))
            if kwargs['draw']:
                draw.line((col, y1, col, y2))

        return white_segments

    def two_longest_runs_on_opp_sides(self, start, end, **kwargs):
        '''Find the two longest runs of white segments in consecutive columns
        that line up.  And make sure the two runs are on opposite
        sides of the page.'''

        comm.verbose("   Finding margins on either side of the page.")
        #comm.debug("start: %s\nend: %s" % (start, end))

        midpoint =  self.im.w / 2
        while len(start) >= 2:
            max, max2 = self.two_longest_runs(start, end)
            if not ((start[max] < midpoint and start[max2] < midpoint) or
                    (end[max] > midpoint and end[max2] > midpoint)):
                return max, max2

            start.remove(start[max2])
            end.remove(end[max2])

        if config.verbose():
            print "max: %d  max2: %d" % (max, max2)
            print "start, end: ", start, end

        return False, False

    def two_longest_runs(self, start, end, **kwargs):
        '''Find the two longest runs of white segments in consecutive columns
        that line up.'''
    
        comm.verbose("   Finding two longest runs.")

        max = 0
        max2 = 0
        if start[0] != -1:
            start.insert(0,-1)
            end.insert(0,-2)
        for r in range(len(start)):
            if end[r] - start[r] > end[max]-start[max]:
                max2 = max
                max = r
            elif end[r] - start[r] > end[max2]-start[max2]:
                max2 = r
        return max, max2

    def get_left_right_crop(self, start, end, max, max2, **kwargs):
        'find left and right crop lines'

        comm.verbose("   Getting left and right crop lines.")

        if start[max] < start[max2]:
            left = end[max]
            right = start[max2]
        else:
            left = end[max2]
            right = start[max]

        return left, right

    def get_top_bottom_crop(self, start, end, max, max2, segments, **kwargs):
        '''Given the starting and ending white space values and the places
        where they run to the end of the page, return top and bottom
        minimum crop rows.'''

        comm.verbose("   Getting top and bottom crop.")

        ## Collect start and end values in the columns of the two runs
        s = []
        e = []
        for x in range(start[max], end[max]):
            s.append(segments[x][0])
            e.append(segments[x][1])
        for x in range(start[max2], end[max2]):
            s.append(segments[x][0])
            e.append(segments[x][1])

        ## Get the top and bottom values from the collected start and end
        start = remove_all_outliers(s)
        end = remove_all_outliers(e)
        top = numpy.max(start)
        bottom = numpy.min(end)

        return top, bottom

    def potential_side_margins(self, segments, **kwargs):
        '''Returns two lists, one with start columns and the other with end
        columns.  The lines those lists describe are consecutive
        columns that are mostly whitespace.  The side margins are
        somewhere in that data.'''

        comm.verbose("   Finding potential side margins.")

        # send a line through.  If it makes it with no collision,
        # we're in the side margin. (This comment might be misleading)
        last_s = (self.im.h, 0)
        last = False
        longest = 0
        diff = self.im.h * 0.1
        start=[]
        end=[]
        for x in range(len(segments)):
            if (abs(segments[x][0] - last_s[0]) < diff and 
                abs(segments[x][1] - last_s[1]) < diff and 
                longest - (segments[x][1] - segments[x][0]) < diff):

                if kwargs['draw']:
                    draw.point((x, self.im.h/3))

                if not last:
                    start.append(x)
                    end.append(0)
                end[-1] = x

                length = segments[x][1] - segments[x][0]
                if length > longest:
                    longest = length

                last = True
            else:
                last = False

            last_s = segments[x]

        return start, end

    def thru_test(self, segments=0, **kwargs):
        '''scan the entire page for white segments and divine left and right
        crops from it.'''

        kwargs = set_defaults( {'draw' : False}, kwargs)

        if kwargs['draw']:
            draw = ImageDraw.Draw(self.im.im)

        segments = self.find_white_segments(
            y_range = self.im.h, y_skip=2, **kwargs)

        start, end = self.potential_side_margins(segments, **kwargs)

        # blank page check
        if len(start) == 1 and len(end) == 1:
            comm.verbose("   Blank page")
            return self.im.w / 2,  self.im.w / 2 + 2

        ## find the longest two runs, these should be the left and right margins
        max, max2 = self.two_longest_runs_on_opp_sides(start, end, **kwargs)

        if max == False and max2 == False:
            comm.verbose("   Blank page")
            return self.im.w / 2,  self.im.w / 2 + 2

        if kwargs['draw']:
            draw.line((start[max], self.im.h/3 - 10, 
                       end[max], self.im.h/3 - 10))
            draw.line((start[max2], self.im.h/3 - 10, 
                       end[max2], self.im.h/3 - 10))

        left, right = self.get_left_right_crop(start, end, max, max2, **kwargs)
        #top, bottom = self.get_top_bottom_crop(
        #    start, end, max, max2, segments, **kwargs)

        if kwargs['draw']:
            draw.line((0, top, self.im.w, top))
            draw.line((0, bottom, self.im.w, bottom))
            draw.line((left, 0, left, self.im.h))
            draw.line((right, 0, right, self.im.h))

        return left,  right

    def shrink_until_white_width(self, edge, **kwargs):
        '''Shrink side of the image until we have a white segment that is >=
        half the height.

        edge must be 'left' or 'right' '''

        assert (edge == 'left' or 
                edge == 'right'), "edge must be 'left' or 'right'"

        kwargs = set_defaults( {'draw' : False}, kwargs)
        if kwargs['draw']:
            draw = ImageDraw.Draw(self.im.im)
        comm.verbose("   Rough shrinking %s edge." % edge)

        col = self.im.w * (edge == 'right')
        y1 = y2 = 0

        if edge == 'right':
            while col > 0 and y2 - y1 < self.im.h / 2:
                col = col - 1
                y1, y2 = self.find_whitest_segment(
                    col, y_skip = 2, y_range = self.im.h, **kwargs)
        else:
            while col < self.im.w and y2 - y1 < self.im.h / 2:
                y1, y2 = self.find_whitest_segment(
                    col, y_skip = 2, y_range = self.im.h, **kwargs)
                col = col + 1

        if kwargs['draw']:
            draw.line((col, 0, col, self.im.h))
        return col

    def shrink_until_white_height(self, edge, **kwargs):
        '''Shrink the top or bottom edge of the image until we have a white
        segment that is >= half the width.  Edge == 'top' or 'bottom'. '''

        assert (edge == 'top' 
                or edge == 'bottom'), "edge must be 'top' or 'bottom'"

        kwargs = set_defaults( {'draw' : False}, kwargs)
        if kwargs['draw']:
            draw = ImageDraw.Draw(self.im.im)
        comm.verbose("   Rough shrinking %s edge." % edge)

        row = self.im.h * (edge == 'bottom')
        y1 = y2 = 0

        if edge == 'top':
            while row < self.im.h and y2 - y1 < self.im.w / 2:
                y1, y2 = self.find_whitest_segment_horiz(row, 
                                                         x_skip = 2, 
                                                         x_range = self.im.w, **kwargs)
                row = row + 1
        else:
            while row > 0 and y2 - y1 < self.im.w / 2:
                row = row - 1
                y1, y2 = self.find_whitest_segment_horiz(row, 
                                                         x_skip = 2, 
                                                         x_range = self.im.w, **kwargs)

        if kwargs['draw']:
            draw.line((0, row, self.im.w, row))
        return row

    def crop(self, **kwargs):
        kwargs = set_defaults( {'draw' : False}, kwargs)

        self.unenhanced = self.im
        self.im.enhance(100,100)
        self.im.getdata()

        ## rough in the edges
        r = self.shrink_until_white_width('right', **kwargs)
        l = self.shrink_until_white_width('left', **kwargs)
        self.im = self.im.crop((l, 0, r, self.im.h))
        t = self.shrink_until_white_height('top', **kwargs)
        b = self.shrink_until_white_height('bottom',**kwargs)
        self.im = self.im.crop((0,t,self.im.w,b))

        # crop tight
        left, right = self.thru_test(**kwargs)
        top = self.lower_the_top(left, 0, right, self.im.h, **kwargs)
        bottom = self.raise_the_bottom(left, top, right, self.im.h, **kwargs)
        left, right = self.tune_sides(left, top, right, bottom, **kwargs)
        crop = self.unenhanced.crop((l+left, t+top, right+l, bottom+t))
        crop.crop = (l+left, t+top, right+l, bottom+t)
        return crop





class SanityCheck(CropMisc):
    '''Do sanity check of a group of cropped scans by comparing the pages
    to the average page and looking for outliers.  We use 2.5 std
    deviations from the mean as our outlier mark, which is a tighter
    boundary than is usually picked.  3 is more common, but that
    leaves too much variance for us.

    M = the data model

    On init, we collect data on our files
    self.find_out_of_spec locates all the bad crops
    self.do_sanity_check recrops bad crops
    '''

    def __init__(self, M, files=[]):
        self.M = M
        self.files=files

        width=[]
        height=[]
        left = []
        right = []
        top = []
        bottom = []
        for c in M.crop.lines:
            w = M.crop.lines[c]['crop'][2] - M.crop.lines[c]['crop'][0]
            h = M.crop.lines[c]['crop'][3] - M.crop.lines[c]['crop'][1]

            # If page isn't blank, add it to the lists
            if w > 25 and h > 25:
                width.append(w)
                height.append(h)

                # abs val of left and right are meaningful becase we
                # line up the spine.  Top and bottom vary with
                # placement of device.
                left.append(M.crop.lines[c]['crop'][0])
                top.append(M.crop.lines[c]['crop'][1])
                right.append(M.crop.lines[c]['crop'][2])
                bottom.append(M.crop.lines[c]['crop'][3])

        ## Make arrays of acceptable dimension values.  Any scans that
        ## fall outside of this set are out of spec.
        remove_all_outliers(width, 2.5)
        remove_all_outliers(height, 2.5)
        remove_all_outliers(left, 2.5)
        remove_all_outliers(right,2.5)
        remove_all_outliers(top, 2.5)
        remove_all_outliers(bottom,2.5)
        self.w_mean = int(numpy.mean(width))
        self.h_mean = int(numpy.mean(height))
        self.left_mean = int(numpy.mean(left))
        self.right_mean = int(numpy.mean(right))
        self.top_mean = int(numpy.mean(top))
        self.bottom_mean = int(numpy.mean(bottom))

        self.width = width
        self.height = height
        self.left = left
        self.right = right

        ## Bring in the subroutines that are common to the original crop class
        CropMisc.__init__(self)

    def find_out_of_spec(self):
        '''Examine M.crop.lines and find out of spec files. Save the
        failed pages in self.fail'''

        fail = {}
        for c in sorted(self.M.crop.lines):
            edge={}
            if self.files == [] or self.M.crop.lines[c]['fname'] in self.files:

                ## If the width is too far from the mean, it is wrong
                w = self.M.crop.lines[c]['crop'][2]-self.M.crop.lines[c]['crop'][0]
                if w > 25 and not w in self.width:
                    if not self.M.crop.lines[c]['crop'][0] in self.left:
                        edge['left'] = True
                    if not self.M.crop.lines[c]['crop'][2] in self.right:
                        edge['right'] = True
                    if not 'left' in edge and not 'right' in edge:
                        edge['width'] = True

                ## If the height is too far from the mean, it is wrong
                h = self.M.crop.lines[c]['crop'][3]-self.M.crop.lines[c]['crop'][1]
                if h > 25 and not (h) in self.height:
                    edge['height']=True
                
                if edge:
                    fail[c] = edge

        self.fail = fail
        return fail

    def _recrop(self, f, c):
        '''Do the image processing to recrop a self.im.  Saves it to
        self.M.crop.lines[f]['dest']

        f is the index in crop_data
        c is new new crop tuple (left, top, right, bottom)

        Called by self._recrop_out_of_spec
        '''

        self.M.crop.lines[f]['crop'] = c
        self.im = self.im.crop(self.M.crop.lines[f]['crop'])
        self.im.save(self.M.crop.lines[f]['dest'])
        comm.verbose("Adjusted %s" % (self.M.crop.lines[f]['dest']))

    def _recrop_out_of_spec(self, f):
        '''Try to fix a messed up file.'''

        crop = list(self.M.crop.lines[f]['crop'])
        width = crop[2] - crop[0]
        height = crop[3] - crop[1]
        fname = self.M.crop.lines[f]['fname']
        recrop_need = False

        if ('width' in self.fail[f] or 'left' in self.fail[f] 
            or 'right' in self.fail[f]):
            self.im = graphic.graphic(self.M.crop.lines[f]['fname'])
            self.im.enhance(100,100)            
                
            l = self.next_white_col(crop[0], crop[1], crop[2], crop[3], 
                                        self.left_mean, 'left', 5)
            r = self.next_white_col(l, crop[1], crop[2], crop[3], 
                                        l + self.w_mean, 'right', 5)
            crop[0], crop[2] = self.breathing_room_sides(l, r)

            crop[1] = self.lower_the_top(crop[0], 0, crop[2], 
                                         self.M.crop.lines[f]['crop'][3])
            crop[3] = self.raise_the_bottom(crop[0], crop[1], 
                                            crop[2], self.im.h)
                    
            recrop_need = True
            for e in ['left', 'right', 'width', 'height']:
                if e in self.fail[f]:
                    del self.fail[f][e]

            
        if 'height' in self.fail[f]:
            ##TODO: Should we deal with out of spec height
            pass

        if recrop_need:
            self._recrop(f, crop)

        #self.M.save()

    def do_sanity_check(self, update_group = None, update_file = None):
        '''Do the sanity check

        update_group and update file = function references.  Pass a
        function that takes an arg.  The func should display the arg
        as text to the user.

        TODO: figure out minimum statistically signif sample.  25 is just a number.
        TODO: each group should have separate stats b/c of camera setup differences'''

        if len(self.M.crop.lines) < 25: 
            return

        update_group("Applying statistical correction to cropped images.")

        fail = self.find_out_of_spec()

        count = 1
        total = len(self.fail)
        for f in sorted(self.fail):
            if os.path.exists('pause'):
                sys.exit()

            update_file(f)
            yield count / (total*1.0)
            self._recrop_out_of_spec(f)
            count += 1
        return
