#!/usr/bin/python

'''Routines for reading, cleaning, and sorting bkrpr input files.'''

import os
import re
from copy import deepcopy
        
import config
import misc

def basename(stub):
    '''Return filespec with the extension (if any) removed.'''

    if stub.rpartition('.')[0]:
        stub = stub.rpartition('.')[0]
    if stub.rpartition('.png')[0]:
        stub = stub.rpartition('.png')[0]
    if stub.rpartition('.tif')[0]:
        stub = stub.rpartition('.tif')[0]
    if stub.rpartition('.crop')[0]:
        stub = stub.rpartition('.crop')[0]
    if stub.rpartition('.rotate')[0]:
        stub = stub.rpartition('.rotate')[0]
    if stub.rpartition('.lines')[0]:
        stub = stub.rpartition('.rotate')[0]
    if stub != '':
        return stub

    return stub


def collect_files(path):
    '''Return a list of files with the complete path.

    path is the dir from which we searched for files

    This func returns a list of the files and the files in the subdir
    with complete paths.  It's not recursive, though.  It only goes
    one level deep.  '''

    ## Collect all files with complete paths
    ls = os.listdir(path)
    files = {}
    files[path] = []
    dirs = []
    for p in ls:
        f = path + '/' + p
        if os.path.isdir(f):
            dirs.append(f)
        elif os.path.isfile(f):
            files[path].append(f)

    for d in dirs:
        files[d] = []
        ls = os.listdir(d)
        for p in ls:
            f = d + '/' + p
            files[d].append(f)
    return files

def clean_file_list(files):
    '''Pare list of files down to image files we can read and process.

    Given the file list (a dict of dirs containing complete-path
    filenames), remove any that do not end in an extension listed in
    config.in_formats or config.in_formats_lc.  Also, remove files we
    generated (*.rotate.png, *.crop.png).'''

    empty_dir = []
    formats = config.in_formats + config.in_formats_lc
    for d in files:
        files_copy = deepcopy(files[d])
        for f in files_copy:
            ext = misc.get_extension(f)
            ext2 = misc.get_extension(f, 2)
            if (not ext in formats) or (ext2 in ['rotate.png', 'crop.png']):
                files[d].remove(f)

        if len(files[d]) == 0:
            empty_dir.append(d)

    for d in empty_dir:
        del files[d]

    return files

def group_files(files):
    '''files - a dict of dirs.  Each val in the dict is a list of
    filespecs with complete paths.

    Return a dict of lists of filespecs with complete paths.  Each
    group has names that consist of text and at least one run of
    numbers.  The text in all the names is the same.  If there are
    multiple runs of numbers, all but the rightmost grouping is
    treated as text (not numbers).  The rightmost grouping is the
    significant one and pages will be ordered according to it.  All the
    other numbers will be the same for each file.

    The name includes the full path, so pictures in different
    directories will be in different groups, even if they have the
    same name/number scheme in the subdirs.
    '''

    group = {}
    pat = re.compile('\d+')
    for d in files:
        for f in files[d]:
            ## Replace runs of numbers with token:
            ## Reverse string and only do first match (i.e. only do
            ## *last* match)
            group_name = pat.sub(config.fname_token, f[::-1], 1)[::-1]
            if not group_name in group:
                group[group_name] = []
            group[group_name].append(f)
    return group

def _test():
    '''Executing this files just does doctest.'''

    print "No doc tests written."

    #import doctest
    #doctest.testmod()

if __name__ == "__main__":
    _test()
