#!/usr/bin/python
'''This module contains functions of generic use and application.'''

import os
import sys
import re
import glob
import numpy

import graphic

def get_extension(fname, segments = 1):
    '''Given a file spec, return everything after the ., if any.

    If segments > 1, return that many segments.  (e.g. if segments = 2,
    foo.bar.baz will return bar.baz) 

    >>> get_extension('foo.bar.baz', 1)
    'baz'
    >>> get_extension('foo.bar.baz', 2)
    'bar.baz'
    >>> get_extension('foobarbaz', 1)
    ''
    >>> get_extension('foobar.baz', 2)
    'baz'
    '''

    period = ''
    out = ''
    for s in range(segments):
        p = fname.rpartition('.')
        fname = p[0]
        if p[1] == '.':
            out = p[2] + period + out
        period = '.'
    return out

def find_files_glob(pathname):
    '''Return a list of files matching the glob.'''
    return glob.glob(pathname)

def find_files_regex(dirname, regex='.'):
    '''Return a list of files in the specified dir matching the regex.
    Excludes subdirs'''
    base, dirs, files = iter(os.walk(dirname)).next() 
    test = re.compile(regex)
    files = filter(test.search, files)
    return files

def syscall(command, capture=False):
    'Do a system call.  If capture is true, returns output.'
    if capture:
        fin = os.popen(command)
        lines = ''
        for line in fin.readlines():
            lines = lines + line
        fin.close()
        return lines
    else:
        os.system(command)

def set_defaults(defaults, kwargs):
    '''For keys that don't exist in kwargs but do in defaults, set to the
    default value.'''
    for f in defaults:
        if not f in kwargs:
            kwargs[f] = defaults[f]
    return kwargs

def slurpfile(filename):
    '''Grab a file's entire contents and return it.'''
    try:
        flh = open(filename,'r')
    except IOError:
        print "IOError: ", sys.exc_value
        return None
    lines = flh.read()
    flh.close()
    return lines

def file2array(filename):
    'Slurp all the lines of a file into an array'

    try:
        flh = open(filename)
    except IOError:
        print "IOError: ", sys.exc_value
        return None

    ret = flh.readlines()
    flh.close()
    return ret


def slope(pt1, pt2):
    'return slope of pt1,pt2'
    dx = float(pt1[0]-pt2[0])
    dy = pt1[1]-pt2[1]
    if dx == 0:
        dx = .000000000000001 # divide by zero
    return dy/dx
def mean(*args):
    'return average of elements in the list'
    total = 0
    for a in args:
        total = total + a
    return float(total) / len(args)

def calc_right_tri_leg(angle, leg):
    '''Given an angle in degrees and the length of the non-corresponding
    leg in a right triangle, return the length of the other leg of
    that triangle.'''

    leg = float(leg)
    angle = float(angle)
    if angle >= 90:
        raise Exception, "A right triangle can't have any angles > 90 degrees." 
    return leg * angle / (90-angle)

def remove_outliers(start, std_dev=3):
    '''Remove statistical outliers from a list of numbers.  Outlier is > std_dev
    std deviations from the mean.  std_dev is an optional arg'''

    for s in start:
        if abs(s - numpy.mean(start)) / numpy.std(start) > std_dev:
            start.remove(s)
    return start

def remove_all_outliers(start, std_dev=3):
    '''Remove outliers over and over until none are left.

    An outlier is an element > std_dev standard deviations from the
    mean.  std_dev is an optional argument.
    '''

    last = 0
    while len(start) != last:
        last = len(start)
        start = remove_outliers(start, std_dev)
    return start

def file_or_image(image):
    '''if image is a string, load the file there as an image.  If it's
    already an image, just return it.'''

    if type(image) == type('str'):
        return graphic.graphic(image)
    if image.__class__.__name__ == 'graphic':
        return image
    if image.__class__.__name__ == 'image':
        return image

    assert False, "Image must be string, graphic or image, not %s" % type(image)

def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()
