#!/usr/bin/python

'''We often want to send one line of text to the user.  This is
classes that allow us to do that in a way that hides from the caller
in what context that communication happens.  If there is a gui, we'll
use the status bar.  If there isn't, we'll print a line on the CLI.
If logging is enabled, we'll do that as needed.  We'll respect verbose
settings too.

This will allow cropper and rotater to print status messages without
knowing GUI/CLI status.

This module provides an *instance* of the one_line class called comm.
"from one_line import comm" will allow you to use it as a global comm
system in your code.  You'll want to
"comm.set_status_bar_active(True)" to get the status bar working.
Alternatively, you can instantiate one_line yourself and pass that
comm system around as needed.  Since this is essentially a smarter
print function, I'm inclined to use the global version.

TODO: Implement logging'''

import logging

class CLI(object):
    '''Class for printing to the commandline.  Yeah, it's weird to need a
    class for that, but we have a class var and we have classes for
    logging and status bar, so this keeps things consistent.'''

    active = True
    def __init__(self):
        'We do not actually do any init for this function.'
        pass

    def out(self, msg):
        'Print a message on the commandline'
        if self.active:
            print msg

    def set_active(self, active):
        'An inactive commandline will ignore requests for output.'
        self.active = active

class OneLine(object):
    '''Central interface for sending one line of text to the status_bar,
    CLI, and/or logger.'''

    last = None
    history = ''

    _verbose = False # set to true for comm.verbose conditional publish
    _debug = False

    def __init__(self):
        '''Create the status bar, the CLI instance and the logger instance.
        Note: logger not yet implemented.'''
        self.status_bar = None
        self.CLI = CLI()

    def set_status_bar_active(self, active):
        if active:
            try:
                import pygtk
                pygtk.require('2.0')
                import gtk
                self.status_bar = gtk.Statusbar()
            except:
                pass

            try:
                self.context_id
            except:
                self.context_id = self.status_bar.get_context_id("Oneline Status Bar")

        try:
            self.status_bar.active=active
        except:
            pass
            

    def clear(self):
        'Clear the statusbar.'
        self.status_bar.set('')

    def out(self, message):
        '''Output a one-line message to some combination of the status bar,
        the CLI and log.'''

        if message != '':
            try:
                if self.status_bar.active:
                    msg_id = self.status_bar.push(self.context_id, message)
                    if msg_id > 1:
                        self.status_bar.remove(self.context_id, msg_id -1)
            except:
                pass
            self.CLI.out(message)
            self.last = message
            self.history = self.history + "\n" + message

    def get_last(self):
        '''Return the last message output via self.out'''
        return self.last

    def set_verbose(self, v):
        self._verbose = v

    def verbose(self, *arg):
        if self._verbose or self.debug:
            self.out(*arg)

    def set_debug(self, v):
        self._debug = v

    def debug(self, *arg):
        if self._debug:
            self.out(*arg)

comm = OneLine()

